;;; -*- lisp -*-

;;; `defpatt' - A package for defining PPCREs for the `cl-ppcre' package
;;;             by Dr. Edi Weitz.

;;; EXAMPLES of `defpatt' use at the end of this file.

;;;
;;; Copyright (c) 2004, Klaus Harbo.   All rights reserved.
;;;

;;; Redistribution and use in source and binary forms, with or without
;;; modification, are permitted provided that the following conditions
;;; are met:

;;;   * Redistributions of source code must retain the above copyright
;;;     notice, this list of conditions and the following disclaimer.

;;;   * Redistributions in binary form must reproduce the above
;;;     copyright notice, this list of conditions and the following
;;;     disclaimer in the documentation and/or other materials
;;;     provided with the distribution.

;;; THIS SOFTWARE IS PROVIDED BY THE AUTHOR 'AS IS' AND ANY EXPRESSED
;;; OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
;;; WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
;;; ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
;;; DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
;;; DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
;;; GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
;;; INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
;;; WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
;;; NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
;;; SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

(in-package :defpatt)

(defvar *defpatt* nil
  "Holds the pattern functions and symbols defined using `defpatt'.")

(defparameter *local-readtable* nil)

(defun %parse-string (str)
  (cl-ppcre::parse-string str))

(defmacro defpatt-pattern (expr)
  "Interprets `expr' according to pattern definitions defined using `defpatt'. ~
     `Defpatt-pattern' is typically not used literally in programs, but invoked ~
     'under the covers' using macro characters (cf. `defpatt-set-macro-char' and ~
     `defpatt-set-dispatch-char')."
  (if (stringp expr)
      `',(%parse-string expr)
      (let ((syms)
	    (funs))
	(maphash (lambda (k v)
		   (push (list k v (gensym)) syms)) (syms *defpatt*))
	(maphash (lambda (k v)
		   (push (cons k v) funs)) (funs *defpatt*))
	`(symbol-macrolet ,(loop for (k  v gensym) in syms
                                 collect (list k `,(list gensym)))
           (labels ,(append
                     (loop for (k . (args . body)) in funs
                       collect (list k args body))
                     (loop for (k  v gensym) in syms
                       collect (list gensym () v)))
	      ,expr)))))

(defun defpatt-reader-function (stream char &optional arg)
  (declare (ignore char arg))
  (let ((expr (read stream t nil t)))
    `(defpatt-pattern ,expr)))

(defun defpatt-set-macro-char (char &optional (readtable *readtable*))
  "Defines `char' as a macro character in `readtable', which interprets the ~
     following expression as a `defpatt' expression.  See also ~
     `defpatt-set-default-macro-char' for setting the default macro character (#\)."
  (set-macro-character char #'defpatt-reader-function readtable))

(defun defpatt-set-default-macro-char (&optional (readtable *readtable*))
  "Sets default macro character for defpatt patterns (#\).  See also ~
     `defpatt-set-macro-char'."
  (defpatt-set-macro-char #\ readtable))

(defun defpatt-set-dispatch-macro-char (dispchar subchar &optional (readtable *readtable*))
  "Sets `dispchar' and `subchar' to be the dispatch macro character pair, which ~
     interprets the following expression as a `defpatt' expression.  Evaluating ~
     `(defpatt-set-dispatch-macro-char #\# #\)' enables `defpatt' expressions to be ~
     entered as (e.g.) `#(upto \"end-of-data\")'."
  (set-dispatch-macro-character dispchar subchar #'defpatt-reader-function readtable))
    
(defclass defpatt ()
  ((syms :initarg :syms
	 :initform nil
	 :accessor syms)
   (funs :initarg :funs
	 :initform nil
	 :accessor funs)))

(defmethod initialize-instance :after ((defpatt defpatt) &rest initargs)
  (setf (syms defpatt) (make-hash-table))
  (setf (funs defpatt) (make-hash-table)))
  
(defmacro defpatt (name args body)
  "Interface to defining `defpatt cl-ppcre pattern abstractions."
  `(progn
    (if ',args
	(setf (gethash ',name (funs *defpatt*)) (cons ',args ',body))
	(setf (gethash ',name (syms *defpatt*)) ',body))
    ',name))

(defun defpatt-initialize-abstractions ()
  ;;;
  ;;; REG/ALT/SEQ - cf. cl-ppcre documentation
  ;;;
  (defpatt REG (&rest args) `(:REGISTER ,@args))
  (defpatt ALT (&rest args) `(:ALTERNATION ,@args))
  (defpatt SEQ (&rest args) `(:SEQUENCE ,@args))

  ;;;
  ;;; GROUP
  ;;;
  (defpatt group (&rest args) `(:GROUP ,@args))

  ;;;
  ;;; REP - Repetition
  ;;;
  (defpatt REP (patt &optional (min 0) (max nil))
    `(:GREEDY-REPETITION ,min ,max ,patt))
  (defpatt ?? (patt) (rep patt 0 1))
  (defpatt ** (patt) (rep patt 0 nil))
  (defpatt ++ (patt) (rep patt 1 nil))

  ;;;
  ;;; RE - Regular Expression String
  ;;;
  (defpatt RE (patt)
    (progn
      (unless (stringp patt)
        (error "Not a string: ~a." patt))
      (%parse-string patt)))

  ;;;
  ;;; WS - White-Space
  ;;;
  (defpatt WS  () :WHITESPACE-CHAR-CLASS)
  (defpatt WS* () (** WS))
  (defpatt WS+ () (++ WS))

  ;;;
  ;;; DIGIT
  ;;;
  (defpatt DIGIT  () '(:CHAR-CLASS (:RANGE #\0 #\9)))
  (defpatt DIGIT* () (** DIGIT))
  (defpatt DIGIT+ () (++ DIGIT))

  ;;;
  ;;; UPTO - Match all data up-to-but-not-including `patt'
  ;;;        (I use this one all the time)
  (defpatt UPTO (patt)
    `(:SEQUENCE (:FLAGS :SINGLE-LINE-MODE-P)
      (:GREEDY-REPETITION 0 NIL 
       (:SEQUENCE 
	:EVERYTHING 
	(:NEGATIVE-LOOKAHEAD ,patt)))
      :EVERYTHING))
  (defpatt UPTO+ (patt)			; include `patt' in match
    `(:SEQUENCE ,(UPTO patt) ,patt))
  ;;;
  ;;; Note
  ;;;
  ;;; The more `defpatt' abstractions are available, the more useful
  ;;; `defpatt' will be.  Contributions are gratefully accepted - 
  ;;; feel free to  e-mail the `defpatt' maintainer suggestions
  ;;; (cf. ./cl-ppcre-defpatt.asd).
  ;;;
  ) ; defpatt-initialize-abstractions

(defun defpatt-initialize (&key (patterns t) (default-macro-char t))
  (setf *defpatt* (make-instance 'defpatt))
  (when patterns
    (defpatt-initialize-abstractions))
  (when default-macro-char
    (defpatt-set-default-macro-char)))


#| EXAMPLES

  ; If you want to try the examples, be sure to evaluate the 
  ; expression below first - otherwise the other ones won't work.

  ; If you want to use a dispatching macro character, it must be
  ; defined separately.

> (defpatt:defpatt-set-default-macro-char)

  ; Note: Introduces #\ as a macro character

=> T

> (cl-ppcre:all-matches-as-strings (alt "a" "c" "f") "abcdefghi")

  ; Note: Equivalent to "a|c|f"

=> ("a" "c" "f")

  ; That's all very well, but doesn't buy us very much.
  ; However `defpatt' (as per cl-ppcre's sexpr-based 
  ; representation of REs) enables us to both document 
  ; the patterns much better by letting us insert comments
  ; into REs...

> (cl-ppcre:scan-to-strings
         (seq digit+  ; used space
	       ws+     
	       digit+  ; available space
	       ws+
	       digit+  ; remaining space 
	       ) "123   4567   7887")

  ; Note: `ws+' and `digit+' are defined above, in `defpatt-initialize'.

=> "123   4567   7887", #()

  ; ...as well as lets us capture data in a structured fashion...

> (cl-ppcre:register-groups-bind (used avail remain)
	     ((seq (reg digit+)  ; used space
		    ws+           
		    (reg digit+)  ; available space
		    ws+
		    (reg digit+)  ; remaining space 
		    ) "123   4567   7887")
	     (mapcar #'parse-integer (list used avail remain)))

  ; Note: `(reg ...)' creates a register binding

=> (123 4567 7887)

  ; ...but also lets us _FIRST_ define and document the abstraction...

> (defpatt match-nums () 
    (seq (reg digit+)  ; used space
         ws+
         (reg digit+)  ; available space
         ws+
         (reg digit+)  ; remaining space 
         ))
=> MATCH-NUMS

  ; ...and _THEN_ use it...

> (cl-ppcre:register-groups-bind (used avail remain)
	     (match-nums "123   4567   7887")
	   (mapcar #'parse-integer (list used avail remain)))

=> (123 4567 7887)

  ; which is a lot more easily understood, as I am sure you will
  ; agree.

> (cl-ppcre:scan-to-strings (upto "efg") "abcdefghi")

=> "abcd", #()

> (cl-ppcre:scan-to-strings (upto+ "efg") "abcdefghi")

=> "abcdefg", #()

  ; To see the raw cl-ppcre expansion of a `defpatt' expression, 
  ; simply enter it:

> match-nums

=> (:SEQUENCE
    (:REGISTER (:GREEDY-REPETITION 1 NIL (:CHAR-CLASS (:RANGE #\0 #\9))))
    (:GREEDY-REPETITION 1 NIL :WHITESPACE-CHAR-CLASS)
    (:REGISTER (:GREEDY-REPETITION 1 NIL (:CHAR-CLASS (:RANGE #\0 #\9))))
    (:GREEDY-REPETITION 1 NIL :WHITESPACE-CHAR-CLASS)
    (:REGISTER (:GREEDY-REPETITION 1 NIL (:CHAR-CLASS (:RANGE #\0 #\9)))))

  ; To see _HOW_ `defpatt' expands an expression use `macroexpand':

> (macroexpand-1 'match-nums)

=> (SYMBOL-MACROLET ((MATCH-NUMS (#:G3523))
                     (WS (#:G3522))
                     (DIGIT (#:G3521))
                     (WS* (#:G3520))
                     (DIGIT* (#:G3519))
                     (WS+ (#:G3518))
                     (DIGIT+ (#:G3517)))
     (LABELS ((GROUP (&REST DEFPATT::ARGS) `(:GROUP ,@DEFPATT::ARGS))
              (SEQ (&REST DEFPATT::ARGS) `(:SEQUENCE ,@DEFPATT::ARGS))
              (UPTO (DEFPATT::PATT)
                `(:SEQUENCE
                  (:FLAGS :SINGLE-LINE-MODE-P)
                  (:GREEDY-REPETITION 0 NIL (:SEQUENCE :EVERYTHING (:NEGATIVE-LOOKAHEAD ,DEFPATT::PATT)))
                  :EVERYTHING))
              (++ (DEFPATT::PATT) (REP DEFPATT::PATT 1 NIL))
              (?? (DEFPATT::PATT) (REP DEFPATT::PATT 0 1))
              (UPTO+ (DEFPATT::PATT) `(:SEQUENCE ,(UPTO DEFPATT::PATT) ,DEFPATT::PATT))
              (ALT (&REST DEFPATT::ARGS) `(:ALTERNATION ,@DEFPATT::ARGS))
              (** (DEFPATT::PATT) (REP DEFPATT::PATT 0 NIL))
              (REG (&REST DEFPATT::ARGS) `(:REGISTER ,@DEFPATT::ARGS))
              (RE (DEFPATT::PATT)
                (PROGN
                  (UNLESS (STRINGP DEFPATT::PATT) (ERROR "Not a string: ~a." DEFPATT::PATT))
                  (DEFPATT::%PARSE-STRING DEFPATT::PATT)))
              (REP (DEFPATT::PATT &OPTIONAL (MIN 0) (MAX NIL))
                `(:GREEDY-REPETITION ,MIN ,MAX ,DEFPATT::PATT))
              (#:G3523 () (SEQ (REG DIGIT+) WS+ (REG DIGIT+) WS+ (REG DIGIT+)))
              (#:G3522 () :WHITESPACE-CHAR-CLASS)
              (#:G3521 () '(:CHAR-CLASS (:RANGE #\0 #\9)))
              (#:G3520 () (** WS))
              (#:G3519 () (** DIGIT))
              (#:G3518 () (++ WS))
              (#:G3517 () (++ DIGIT)))
       MATCH-NUMS))

  ; `upto' and `upto+' are good examples of how having an abstraction
  ; mechanism helps keep maintainable and understandable REs.  See
  ; their definitions above.

|#